
/* Chrysalide - Outil d'analyse de fichiers binaires
 * panel.c - équivalent Python du fichier "gui/panels/panel.c"
 *
 * Copyright (C) 2019-2020 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "panel.h"


#include <pygobject.h>


#include <i18n.h>
#include <core/params.h>
#include <gui/panel-int.h>
#include <gui/core/panels.h>
#include <plugins/dt.h>


#include "constants.h"
#include "item.h"
#include "../access.h"
#include "../helpers.h"
#include "../glibext/named.h"
#include "../gtkext/dockable.h"



/* ------------------------ GLUE POUR CREATION DEPUIS PYTHON ------------------------ */


/* Accompagne la création d'une instance dérivée en Python. */
static PyObject *py_panel_item_new(PyTypeObject *, PyObject *, PyObject *);

/* Initialise la classe des panneaux pour l'interface graphique. */
static void py_panel_item_init_gclass(GPanelItemClass *, gpointer);

/* Initialise une instance sur la base du dérivé de GObject. */
static int py_panel_item_init(PyObject *self, PyObject *args, PyObject *kwds);



/* -------------------------- FONCTIONNALITES D'UN PANNEAU -------------------------- */


/* Place un panneau dans l'ensemble affiché. */
static PyObject *py_panel_item_dock(PyObject *, PyObject *);

/* Supprime un panneau de l'ensemble affiché. */
static PyObject *py_panel_item_undock(PyObject *, PyObject *);

/* Fournit le chemin d'accès à utiliser pour les encapsulations. */
static PyObject *py_panel_item_get_named_widget(PyObject *, void *);

/* Fournit une indication sur la personnalité du panneau. */
static PyObject *py_panel_item_get_personality(PyObject *, void *);

/* Fournit le chemin d'accès à utiliser pour les encapsulations. */
static PyObject *py_panel_item_get_path(PyObject *, void *);

/* Définit le chemin d'accès à utiliser pour les encapsulations. */
static int py_panel_item_set_path(PyObject *, PyObject *, void *);

/* Indique si le composant repose sur un support de l'éditeur. */
static PyObject *py_panel_item_get_docked(PyObject *, void *);



/* ---------------------------------------------------------------------------------- */
/*                          GLUE POUR CREATION DEPUIS PYTHON                          */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : type = type du nouvel objet à mettre en place.               *
*                args = éventuelle liste d'arguments.                         *
*                kwds = éventuel dictionnaire de valeurs mises à disposition. *
*                                                                             *
*  Description : Accompagne la création d'une instance dérivée en Python.     *
*                                                                             *
*  Retour      : Nouvel objet Python mis en place ou NULL en cas d'échec.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    PyObject *result;                       /* Objet à retourner           */
    PyTypeObject *base;                     /* Type de base à dériver      */
    bool first_time;                        /* Evite les multiples passages*/
    GType gtype;                            /* Nouveau type de processeur  */
    bool status;                            /* Bilan d'un enregistrement   */

    /* Validations diverses */

    base = get_python_panel_item_type();

    if (type == base)
    {
        result = NULL;
        PyErr_Format(PyExc_RuntimeError, _("%s is an abstract class"), type->tp_name);
        goto exit;
    }

    /* Mise en place d'un type dédié */

    first_time = (g_type_from_name(type->tp_name) == 0);

    gtype = build_dynamic_type(G_TYPE_PANEL_ITEM, type->tp_name,
                               (GClassInitFunc)py_panel_item_init_gclass, NULL, NULL);

    if (first_time)
    {
        status = register_class_for_dynamic_pygobject(gtype, type, base);

        if (!status)
        {
            result = NULL;
            goto exit;
        }

    }

    /* On crée, et on laisse ensuite la main à PyGObject_Type.tp_init() */

    result = PyType_GenericNew(type, args, kwds);

 exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : class  = classe à initialiser.                               *
*                unused = données non utilisées ici.                          *
*                                                                             *
*  Description : Initialise la classe des panneaux pour l'interface graphique.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void py_panel_item_init_gclass(GPanelItemClass *class, gpointer unused)
{
    py_editor_item_init_gclass(G_EDITOR_ITEM_CLASS(class), NULL);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = objet à initialiser (théoriquement).                  *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Initialise une instance sur la base du dérivé de GObject.    *
*                                                                             *
*  Retour      : 0.                                                           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_panel_item_init(PyObject *self, PyObject *args, PyObject *kwds)
{
    PanelItemPersonality personality;       /* Nature du panneau           */
    GNamedWidget *widget;                   /* Composant visuel du panneau */
    int startup;                            /* Recommandation au démarrage */
    const char *path;                       /* Placement à l'affichage     */
    int ret;                                /* Bilan de lecture des args.  */
    GPanelItem *panel;                      /* Panneau à manipuler         */

    static char *kwlist[] = { "personality", "widget", "dock", "path", NULL };

#define PANEL_ITEM_DOC                                                          \
    "PanelItem is an abstract class for panels available in the main GUI"       \
    " interface.\n"                                                             \
    "\n"                                                                        \
    "Instances can be created using the following constructor:\n"               \
    "\n"                                                                        \
    "    PanelItem(personality, widget, dock, path)"                            \
    "\n"                                                                        \
    "Where:\n"                                                                  \
    "* personality defines how many instances of the panels can be created"     \
    " at the same time.\n"                                                      \
    "* widget is an implementation of the pychrysalide.glibext.NamedWidget"     \
    " interface (see pychrysalide.gtkext.BuiltNamedWidget for a ready-to-use"   \
    " helper).\n"                                                               \
    "* dock defines if the panel is docked by default.\n"                       \
    "* path states the location of the panel inside the main GUI.\n"            \
    "\n"                                                                        \
    "The easiest way to pass all this information is probably to use a"         \
    " transitional dictionary inside the panel *__init__()* constructor:\n"     \
    "\n"                                                                        \
    "    params = {\n"                                                          \
    "        'personality' : PanelItemPersonality.PIP_SINGLETON,\n"             \
    "        'widget' : my_named_widget_instance,\n"                            \
    "        'dock' : True,\n"                                                  \
    "        'path' : 'MES'\n"                                                  \
    "    }\n"                                                                   \
    "    super(MyPanel, self).__init__(**params)\n"                             \
    "\n"                                                                        \
    "The PanelItem definition handles internally the supply of the *_widget*"   \
    " attribute for pychrysalide.gui.EditorItem.\n"                             \
    "\n"                                                                        \
    "For more details about the panel path, please refer to"                    \
    " pychrysalide.gtkext.GtkDockable."

    /* Récupération des paramètres */

    ret = PyArg_ParseTupleAndKeywords(args, kwds, "O&O&ps", kwlist,
                                      convert_to_panel_item_personality, &personality,
                                      convert_to_named_widget, &widget,
                                      &startup, &path);
    if (!ret) return -1;

    /* Initialisation d'un objet GLib */

    ret = forward_pygobjet_init(self);
    if (ret == -1) return -1;

    /* Eléments de base */

    panel = G_PANEL_ITEM(pygobject_get(self));

    panel->personality = personality;

    /**
     * Si Python ne voit plus la variable représentant le panneau utilisée,
     * il va la supprimer, ce qui va supprimer le composant GTK.
     *
     * On sera donc en situation de Use-After-Free, dont les conséquences
     * arrivent très vite.
     */
    panel->widget = widget;
    g_object_ref(G_OBJECT(widget));

    panel->dock_at_startup = startup;
    panel->path = strdup(path);

    return 0;

}



/* ---------------------------------------------------------------------------------- */
/*                            FONCTIONNALITES D'UN PANNEAU                            */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Place un panneau dans l'ensemble affiché.                    *
*                                                                             *
*  Retour      : Py_None.                                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_dock(PyObject *self, PyObject *args)
{
    GPanelItem *item;                       /* Panneau à manipuler         */

#define PANEL_ITEM_DOCK_METHOD PYTHON_METHOD_DEF    \
(                                                   \
    dock, "$self, /",                               \
    METH_NOARGS, py_panel_item,                     \
    "Display the panel item in the right place."    \
)

    item = G_PANEL_ITEM(pygobject_get(self));

    g_panel_item_dock(item);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Supprime un panneau de l'ensemble affiché.                   *
*                                                                             *
*  Retour      : Py_None.                                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_undock(PyObject *self, PyObject *args)
{
    GPanelItem *item;                       /* Panneau à manipuler         */

#define PANEL_ITEM_UNDOCK_METHOD PYTHON_METHOD_DEF  \
(                                                   \
    undock, "$self, /",                             \
    METH_NOARGS, py_panel_item,                     \
    "Hide the panel item from the main interface."  \
)

    item = G_PANEL_ITEM(pygobject_get(self));

    g_panel_item_undock(item);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit le chemin d'accès à utiliser pour les encapsulations.*
*                                                                             *
*  Retour      : Chemin d'accès défini.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_get_named_widget(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GPanelItem *item;                       /* Panneau à consulter         */
    GNamedWidget *widget;                   /* Composant nommé à transférer*/

#define PANEL_ITEM_NAMED_WIDGET_ATTRIB PYTHON_GET_DEF_FULL  \
(                                                           \
    named_widget, py_panel_item,                            \
    "Named widget as core component of the panel item."     \
)

    item = G_PANEL_ITEM(pygobject_get(self));
    widget = gtk_panel_item_get_named_widget(item);

    result = pygobject_new(G_OBJECT(widget));

    g_object_unref(G_OBJECT(widget));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit une indication sur la personnalité du panneau.       *
*                                                                             *
*  Retour      : Identifiant lié à la nature du panneau.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_get_personality(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GPanelItem *item;                       /* Panneau à consulter         */
    PanelItemPersonality personality;       /* Valeur native à convertir   */

#define PANEL_ITEM_PERSONALITY_ATTRIB PYTHON_GET_DEF_FULL       \
(                                                               \
    personality, py_panel_item,                                 \
    "Rule for handling panel creations, as a"                   \
    " pychrysalide.gui.PanelItem.PanelItemPersonality value."   \
)

    item = G_PANEL_ITEM(pygobject_get(self));
    personality = gtk_panel_item_get_personality(item);

    result = cast_with_constants_group_from_type(get_python_panel_item_type(), "PanelItemPersonality", personality);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit le chemin d'accès à utiliser pour les encapsulations.*
*                                                                             *
*  Retour      : Chemin d'accès défini.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_get_path(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GPanelItem *item;                       /* Panneau à consulter         */
    const char *path;                       /* Chemin d'accès courant      */

#define PANEL_ITEM_PATH_ATTRIB PYTHON_GETSET_DEF_FULL       \
(                                                           \
    path, py_panel_item,                                    \
    "Get or define the current path of the panel item."     \
)

    item = G_PANEL_ITEM(pygobject_get(self));
    path = gtk_panel_item_get_path(item);

    result = PyUnicode_FromString(path);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                value   = valeur fournie à intégrer ou prendre en compte.    *
*                closure = adresse non utilisée ici.                          *
*                                                                             *
*  Description : Définit le chemin d'accès à utiliser pour les encapsulations.*
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static int py_panel_item_set_path(PyObject *self, PyObject *value, void *closure)
{
    GPanelItem *item;                       /* Panneau à manipuler         */
    const char *path;                       /* Nouveau chemin d'accès      */

    if (!PyUnicode_Check(value))
        return -1;

    item = G_PANEL_ITEM(pygobject_get(self));

    path = PyUnicode_DATA(value);

    gtk_panel_item_set_path(item, path);

    return 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Indique si le composant repose sur un support de l'éditeur.  *
*                                                                             *
*  Retour      : True si le composant est bien incrusté quelque part.         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_panel_item_get_docked(PyObject *self, void *closure)
{
    PyObject *result;                       /* Valeur à retourner          */
    GPanelItem *item;                       /* Panneau à consulter         */
    bool docked;                            /* Statut de l'ancrage         */

#define PANEL_ITEM_DOCKED_ATTRIB PYTHON_GET_DEF_FULL    \
(                                                       \
    docked, py_panel_item,                              \
    "Dock status of the panel item."                    \
)

    item = G_PANEL_ITEM(pygobject_get(self));
    docked = g_panel_item_is_docked(item);

    result = docked ? Py_True : Py_False;
    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_panel_item_type(void)
{
    static PyMethodDef py_panel_item_methods[] = {
        PANEL_ITEM_DOCK_METHOD,
        PANEL_ITEM_UNDOCK_METHOD,
        { NULL }
    };

    static PyGetSetDef py_panel_item_getseters[] = {
        PANEL_ITEM_NAMED_WIDGET_ATTRIB,
        PANEL_ITEM_PERSONALITY_ATTRIB,
        PANEL_ITEM_PATH_ATTRIB,
        PANEL_ITEM_DOCKED_ATTRIB,
        { NULL }
    };

    static PyTypeObject py_panel_item_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.gui.PanelItem",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = PANEL_ITEM_DOC,

        .tp_methods     = py_panel_item_methods,
        .tp_getset      = py_panel_item_getseters,

        .tp_init        = py_panel_item_init,
        .tp_new         = py_panel_item_new,

    };

    return &py_panel_item_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.gui.panels.PanelItem'. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool ensure_python_panel_item_is_registered(void)
{
    PyTypeObject *type;                     /* Type Python 'LoadedBinary'  */
    PyObject *module;                       /* Module à recompléter        */
    PyObject *dict;                         /* Dictionnaire du module      */

    type = get_python_panel_item_type();

    if (!PyType_HasFeature(type, Py_TPFLAGS_READY))
    {
        module = get_access_to_python_module("pychrysalide.gui");

        dict = PyModule_GetDict(module);

        if (!ensure_python_editor_item_is_registered())
            return false;

        if (!ensure_python_dockable_is_registered())
            return false;

        if (!_register_class_for_pygobject(dict, G_TYPE_PANEL_ITEM, type,
                                           get_python_editor_item_type(), get_python_dockable_type(), NULL))
            return false;

        if (!define_panel_item_constants(type))
            return false;

    }

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : arg = argument quelconque à tenter de convertir.             *
*                dst = destination des valeurs récupérées en cas de succès.   *
*                                                                             *
*  Description : Tente de convertir en panneau pour GUI.                      *
*                                                                             *
*  Retour      : Bilan de l'opération, voire indications supplémentaires.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

int convert_to_panel_item(PyObject *arg, void *dst)
{
    int result;                             /* Bilan à retourner           */

    result = PyObject_IsInstance(arg, (PyObject *)get_python_panel_item_type());

    switch (result)
    {
        case -1:
            /* L'exception est déjà fixée par Python */
            result = 0;
            break;

        case 0:
            PyErr_SetString(PyExc_TypeError, "unable to convert the provided argument to executable format");
            break;

        case 1:
            *((GPanelItem **)dst) = G_PANEL_ITEM(pygobject_get(arg));
            break;

        default:
            assert(false);
            break;

    }

    return result;

}
